#ifndef VECTOR3_H
#define VECTOR3_H

#include <cassert>
#include <cmath>
#include <sstream>

class vector3
{
public:
    const static vector3 ZERO;
    const static vector3 X;
    const static vector3 Y;
    const static vector3 Z;

    vector3()
    {
        this->set(ZERO);
    }

    vector3(double x, double y, double z)
    {
        this->u[0] = x; this->u[1] = y; this->u[2] = z;
    }

    vector3(double v[3])
    {
        this->set(v);
    }

    vector3(const vector3& v)
    {
        this->set(v.u);
    }

    vector3& operator=(const vector3& v)
    {
        this->set(v);
        return *this;
    }
    
    bool operator==(const vector3& v) const
    {
        return this->equal(v);
    }

    bool operator!=(const vector3& v) const
    {
        return !this->equal(v);
    }
    
    vector3 operator+(const vector3& v) const
    {
        return this->add(v);
    }

    vector3 operator-(const vector3& v) const
    {
        return this->sub(v);
    }

    vector3 operator*(double d) const
    {
        return this->mult(d);
    }
    
    bool equal(const vector3& v) const
    {
        if (this->u[0] == v[0] && this->u[1] == v[1] && this->u[2] == v[2])
            return true;
        else
            return false;
    }
    
    double set(int idx, double val)
    {
        assert(idx >= 0 && idx < 3);
        this->u[idx] = val;
    }
    
    void set(const double v[3])
    {
        for (int i = 0; i < 3; ++i)
            this->u[i] = v[i];
    }

    void set(const vector3& v)
    {
        this->set(v.u);
    }

    double get(int idx) const
    {
        assert(idx >= 0 && idx < 3);
        return this->u[idx];
    }

    double& operator[](int idx)
    {
        assert(idx >= 0 && idx < 3);
        return this->u[idx];
    }
    
    const double& operator[](int idx) const
    {
        assert(idx >= 0 && idx < 3);
        return this->u[idx];
    }
    
    vector3 negate() const
    {
        return this->mult(-1.0);
    }

    vector3 add(const vector3& v) const
    {
        return vector3(
            this->u[0] + v[0], this->u[1] + v[1], this->u[2] + v[2]
        );
    }

    vector3 sub(const vector3& v) const
    {
        return this->add(v.negate());
    }
   
    // scalar multiplication 
    vector3 mult(double s) const
    {
        return vector3(this->u[0] * s, this->u[1] * s, this->u[2] * s);
    }

    // component-wise multiplication
    vector3 component_mult(const vector3& v) const
    {
        return vector3(
            this->u[0] * v[0], this->u[1] * v[1], this->u[2] * v[2]
        );
    }

    double dot(const vector3& v) const
    {
        double dot_val = 0.0;
        for (int i = 0; i < 3; ++i)
            dot_val += this->u[i] * v[i];

        return dot_val;
    }

    double length() const 
    {
        double dist2 = 0.0;
    
        for (int i = 0; i < 3; ++i)
            dist2 += pow(this->u[i], 2);

        return sqrt(dist2);
    }

    vector3 normalize() const
    {
        double length = this->length();
        
        assert(length != 0.0);
        
        vector3 v(*this);
        for (int i = 0; i < 3; ++i)
           v[i] /= length; 

        return v;
    }

    vector3 cross(const vector3& v) const
    {
        vector3 u_cross_v;

        u_cross_v[0] = this->u[1] * v[2] - this->u[2] * v[1];
        u_cross_v[1] = this->u[2] * v[0] - this->u[0] * v[2];
        u_cross_v[2] = this->u[0] * v[1] - this->u[1] * v[0];

        return u_cross_v;
    }

    std::string to_string() const
    {
        std::stringstream ss;

        ss << "(" << this->u[0] << ", " << this->u[1] << ", " 
            << this->u[2] << ")";
    
        return ss.str();
    }
    
private:
    double u[3];

}; // class vector3

#endif
